#include <iostream>
#include <filesystem>
#include <fstream>
#include <unordered_map>
#include <vector>
#include <string>
#include <map>
#include <functional>

/*
Autor: Mikołaj Sendybył

* Opis Projektu:
* Projekt implementuje klase FileManager, ktora oferuje interfejs do zarzadzania plikami i katalogami oraz wyszukiwania plikow na podstawie ich rozszerzen.
* Klasa wykorzystuje mozliwosci biblioteki std::filesystem, wprowadzonej w standardzie C++17, aby wykonywac rozne operacje na systemie plikow.
*
* Wprowadzenie do std::filesystem:
*  Biblioteka <filesystem> zostala wprowadzona w C++17 i dostarcza funkcje do wykonywania operacji na systemach plikow oraz ich komponentach,
* takich jak sciezki, pliki regularne i katalogi. Oferuje funkcje do tworzenia, usuwania, kopiowania i zmieniania nazw plikow oraz katalogow,
* a takze do zapytania o ich wlasciwosci. Celem tej biblioteki jest standaryzacja operacji na systemie plikow oraz uczynienie ich bardziej przenosnymi miedzy roznymi platformami.
*
* Zrodla:
* Dokumentacja C++17: https://en.cppreference.com/w/cpp/filesystem
* GeeksforGeeks - Strona oferujaca roznorodne artykuly i przewodniki dotyczace C++: https://www.geeksforgeeks.org/file-system-library-in-cpp-17/
* Dokumentacja Microsoftu dotyczaca std::filesystem: https://learn.microsoft.com/en-us/cpp/standard-library/filesystem?view=msvc-170
*/
namespace fs = std::filesystem;

/**
 * @class FileManager
 * @brief Zapewnia interfejs do zarzadzania plikami i katalogami, w tym wyszukiwanie plikow wedlug rozszerzen.
 */
class FileManager {
public:
    FileManager() {
        initializeCommands_directories();
        initializeCommands_menu();
        initalizeTXT_main_menu();
        initalizeTXT_directories_menu();
    }

    /**
     * @brief Uruchamia glowna petle aplikacji. Wyswietla glowne menu i przetwarza wejscie uzytkownika, aby nawigowac miedzy roznymi funkcjonalnosciami menedzera plikow.
     */
    void run() {
        std::string command;
        do {
            show_menu(menu_command_strings);
            command = user_input_command();
        } while (execute_command(command, menu_command_functions));
    }

private:
    std::map<std::string, std::function<bool()>> directory_command_functions;
    std::map<std::string, std::function<bool()>> menu_command_functions;
    std::vector<std::string> menu_command_strings;
    std::vector<std::string> directory_command_strings;

    /**
     * @brief Inicjalizuje funkcje polecen zarzadzania katalogami.
     */
    void initializeCommands_directories() {
        directory_command_functions["1"] = [this]() { list_directory(); return true; };
        directory_command_functions["2"] = [this]() { create_directory(); return true; };
        directory_command_functions["3"] = [this]() { remove_directory(); return true; };
        directory_command_functions["4"] = [this]() { move_directory(); return true; };
        directory_command_functions["5"] = [this]() { copy_directory(); return true; };
        directory_command_functions["6"] = [this]() { return false; }; // false przerywa petle
    }

    /**
     * @brief Inicjalizuje funkcje polecen glownego menu.
     */
    void initializeCommands_menu() {
        menu_command_functions["1"] = [this]() { run_directories_manager(); return true; };
        menu_command_functions["2"] = [this]() { search_files(); return true; };
        menu_command_functions["3"] = [this]() { return false; };
    }

    /**
     * @brief Inicjalizuje tekst glownego menu.
     */
    void initalizeTXT_main_menu() {
        menu_command_strings = {
            "  1: Manage Directories",
            "  2: Search directory for Files with given extension",
            "  3: Exit"
        };
    }

    /**
     * @brief Inicjalizuje tekst menu zarzadzania katalogami.
     */
    void initalizeTXT_directories_menu() {
        directory_command_strings = {
             "  1. list        - List content of a directory",
             "  2. create_dir  - Create a new directory",
             "  3. remove_dir  - Remove a directory",
             "  4. move        - Move content of a directory",
             "  5. copy        - Copy content of a directory",
             "  6. back        - Return to the main menu"
        };
    }

    /**
     * @brief Wyswietla dostepne polecenia menu.
     * @param menu_commands Wektor stringow reprezentujacych polecenia menu.
     */
    void show_menu(const std::vector<std::string>& menu_commands) {
        std::cout << "Available commands:\n";
        for (const auto& command : menu_commands) {
            std::cout << "- " << command << "\n";
        }
    }

    /**
     * @brief Wykonuje polecenie odpowiadajace wejsciu uzytkownika.
     * @param command Polecenie wprowadzone przez uzytkownika.
     * @param command_functions Mapa polecen i odpowiadajacych im funkcji.
     * @return True jesli petla powinna kontynuowac, false w przeciwnym razie.
     */
    bool execute_command(const std::string& command, const std::map<std::string, std::function<bool()>>& command_functions) {
        auto it = command_functions.find(command);
        if (it != command_functions.end()) {
            return it->second();
        }
        else {
            std::cout << "Unknown command.\n";
            return true;
        }
    }

    /**
     * @brief Zarzadza operacjami na plikach i katalogach. Wyswietla menu zarzadzania katalogami i przetwarza polecenia uzytkownika zwiazane z operacjami na plikach i katalogach.
     */
    void run_directories_manager() {
        std::string command;
        do {
            show_menu(directory_command_strings);
            command = user_input_command();
        } while (execute_command(command, directory_command_functions));
    }

    /**
     * @brief Wyswietla zawartosc katalogu podanego przez uzytkownika.
     * @details
     * fs::path to klasa, ktora pozwala na przenosne zarzadzanie sciezkami systemow plikow w programach.
     *
     * fs::directory_iterator to klasa umozliwiajaca iterowanie po zawartosci katalogow
     * w systemie plikow. Dzieki niej mozna przegladac pliki i podkatalogi znajdujace sie w okreslonym katalogu.
     *
     * fs::directory_entry to klasa z biblioteki, ktora reprezentuje pojedynczy wpis w katalogu, taki jak plik lub podkatalog.
     * Ta klasa jest czesto uzywana w polaczeniu z fs::directory_iterator do iteracji przez zawartosc katalogu,
     * umozliwiajac dostep do informacji o kazdym wpisie, takich jak sciezka, status i atrybuty pliku.
     *
     * Funkcja fs::exists jest uzywana do sprawdzania, czy okreslona sciezka w systemie plikow istnieje.
     * Zwraca true, jesli sciezka istnieje, niezaleznie od tego, czy jest to katalog, plik regularny,
     * czy inne urzadzenie. Zwraca false, jesli sciezka nie istnieje lub wystapil blad podczas proby dostepu do sciezki.
     *
     * Funkcja fs::is_directory jest wykorzystywana do sprawdzania, czy dana sciezka w systemie plikow odnosi sie do katalogu.
     * Funkcja zwraca true jesli podana sciezka jest katalogiem. W przeciwnym wypadku, w tym gdy sciezka nie istnieje lub wystepuje problem z dostepem, zwraca false.
     */
    void list_directory() {
        fs::path path;
        std::cout << "Enter directory path to list: ";
        std::cin >> path;
        if (fs::exists(path) && fs::is_directory(path)) {
            for (const auto& entry : fs::directory_iterator(path)) {
                std::cout << entry.path().filename() << '\n';
            }
        }
        else {
            std::cout << "Directory does not exist.\n";
        }
    }

    /**
     * @brief Tworzy nowy katalog w lokalizacji podanej przez uzytkownika.
     * @details
     * Funkcja fs::create_directories jest jedna z funkcji biblioteki <filesystem>, ktora umozliwia tworzenie katalogow w systemie plikow.
     * Argumenty:
     * p: Sciezka do katalogu, ktory ma zostac utworzony.
     * Jezeli sciezka zawiera kilka poziomow katalogow, ktore jeszcze nie istnieja, fs::create_directories automatycznie tworzy wszystkie potrzebne katalogi nadrzedne.
     * Zwraca True jesli katalog zostal utworzony, False w przeciwnym wypadku.
     */
    void create_directory() {
        fs::path full_path = user_input_remove_create();
        if (!fs::create_directories(full_path)) {
            std::cout << "Failed to create directory or it already exists: " << full_path << '\n';
        }
        else {
            std::cout << "Directory created: " << full_path << '\n';
        }
    }

    /**
     * @brief Usuwa katalog podany przez uzytkownika, wlacznie z jego zawartoscia.
     * @details
     * Funkcja fs::remove_all umozliwia usuwanie katalogu oraz wszystkich jego zawartosci, wlacznie z podkatalogami i plikami.
     * Przydatna w przypadkach, gdy potrzebujemy wyczyscic cale drzewo katalogow - dzialanie jest rekurencyjne.
     * Argumenty:
     * p: Sciezka do katalogu, ktory ma zostac usuniety wraz z cala jego zawartoscia.
     *
     * Funkcja zwraca liczbe usunietych plikow i katalogow;
     * Moze rzucic wyjatek std::filesystem::filesystem_error, jesli napotka problemy z dostepem do sciezki, uprawnieniami lub innymi bledami systemu plikow.
     */
    void remove_directory() {
        fs::path full_path = user_input_remove_create();
        if (fs::remove_all(full_path) == 0) {
            std::cout << "Failed to remove or directory does not exist: " << full_path << '\n';
        }
        else {
            std::cout << "Directory removed: " << full_path << '\n';
        }
    }

    /**
     * @brief Przenosi zawartosc katalogu z lokalizacji zrodlowej do docelowej podanej przez uzytkownika.
     * @details
     * Funkcja fs::rename sluzy do zmiany nazwy pliku lub katalogu lub do przenoszenia pliku lub katalogu do nowej lokalizacji.
     * Przyjmuje dwa argumenty - from: Sciezka zrodlowa pliku lub katalogu, ktory ma byc zmieniony lub przeniesiony.
     * to: Sciezka docelowa, do ktorej plik lub katalog ma byc przeniesiony lub gdzie ma byc zmieniona jego nazwa.
     * Podobnie jak fs::remove_all, fs::rename moze rzucic wyjatek std::filesystem::filesystem_error, jesli wystapia problemy z dostepem do sciezki.
     */
    void move_directory() {
        std::vector<fs::path> out = user_input_move_copy();
        try {
            fs::rename(out[0], out[1]);
            std::cout << "Successfully moved.\n";
        }
        catch (const fs::filesystem_error& e) {
            std::cout << "Error moving: " << e.what() << '\n';
        }
    }

    /**
     * @brief Kopiuje zawartosc katalogu z lokalizacji zrodlowej do docelowej podanej przez uzytkownika.
     * @details
     * Funkcja fs::copy z biblioteki <filesystem> umozliwia kopiowanie plikow i katalogow z jednej lokalizacji do drugiej.
     * Argumenty:
     * from: Sciezka zrodlowa pliku lub katalogu, ktory ma byc skopiowany.
     * to: Sciezka docelowa, gdzie ma zostac umieszczona kopia.
     * options: Opcje kopiowania, ktore okreslaja, jak operacja powinna byc przeprowadzona. Przykladem jest fs::copy_options::recursive uzyta w zadaniu,
     * ktory pozwala na rekursywne kopiowanie katalogow wraz z ich zawartoscia.
     */
    void copy_directory() {
        std::vector<fs::path> out = user_input_move_copy();
        try {
            fs::copy(out[0], out[1], fs::copy_options::recursive);
            std::cout << "Successfully copied.\n";
        }
        catch (const fs::filesystem_error& e) {
            std::cout << "Error copying: " << e.what() << '\n';
        }
    }

    /**
     * @brief Wyszukuje pliki o okreslonych rozszerzeniach w katalogu podanym przez uzytkownika.
     */
    void search_files() {
        fs::path dir_to_scan = user_input_scan();
        std::vector<std::string> file_types = user_input_file_type();
        bool recursive = user_input_recursive();

        std::unordered_map<std::string, int> file_count = initialize_file_count(file_types);
        if (recursive) {
            scan_directory_recursive(dir_to_scan, file_count);
        }
        else {
            scan_directory(dir_to_scan, file_count);
        }
        print_file_count(file_count);
    }

    /**
     * @brief Inicjalizuje mape do zliczania plikow wedlug ich rozszerzen.
     * @param file_types Wektor rozszerzen plikow do zliczenia.
     * @return Unordered_map z rozszerzeniami plikow jako kluczami i ich licznikami jako wartosciami.
     */
    std::unordered_map<std::string, int> initialize_file_count(const std::vector<std::string>& file_types) {
        std::unordered_map<std::string, int> file_count;
        for (const auto& ext : file_types) {
            file_count[ext] = 0; // Initialize count for each extension
        }
        return file_count;
    }

    /**
     * @brief Przeszukuje katalog i zlicza pliki wedlug ich rozszerzen.
     * @param dir_to_scan Katalog do przeszukania.
     * @param file_count Unordered_map do przechowywania liczby plikow wedlug rozszerzen.
     * @details
     *
     * Klasa fs::recursive_directory_iterator umozliwia glebsze przeszukiwanie katalogow, umozliwiajac iteracje nie tylko przez zawartosc bezposredniego katalogu (jak fs::directory_iterator),
     * ale rowniez przez wszystkie jego podkatalogi.
     * Argumenty:
     * p: Sciezka do katalogu, od ktorego ma rozpoczac sie iteracja.
     * options: Opcje okreslajace zachowanie iteratora, np. czy ignorowac bledy dostepu.
     * ec: Opcjonalny parametr do przechwytywania bledow bez rzucania wyjatkow, w przypadku wystapienia problemow podczas tworzenia iteratora.
     */
    void scan_directory_recursive(const fs::path& dir_to_scan, std::unordered_map<std::string, int>& file_count) {
        for (const auto& entry : fs::recursive_directory_iterator(dir_to_scan)) {
            if (entry.is_regular_file()) {
                auto file_ext = entry.path().extension().string();
                if (file_count.find(file_ext) != file_count.end()) {
                    file_count[file_ext]++;
                }
            }
        }
    }

    /**
     * @brief Przeszukuje tylko podany katalog i zlicza pliki wedlug ich rozszerzen.
     * @param dir_to_scan Katalog do przeszukania.
     * @param file_count Unordered_map do przechowywania liczby plikow wedlug rozszerzen.
     */
    void scan_directory(const fs::path& dir_to_scan, std::unordered_map<std::string, int>& file_count) {
        for (const auto& entry : fs::directory_iterator(dir_to_scan)) {
            if (entry.is_regular_file()) {
                auto file_ext = entry.path().extension().string();
                if (file_count.find(file_ext) != file_count.end()) {
                    file_count[file_ext]++;
                }
            }
        }
    }

    /**
     * @brief Wyswietla liczbe znalezionych plikow wedlug ich rozszerzen.
     * @param file_count Unordered_map z rozszerzeniami plikow jako kluczami i ich licznikami jako wartosciami.
     */
    void print_file_count(const std::unordered_map<std::string, int>& file_count) {
        std::cout << "Found files by extension:" << std::endl;
        for (const auto& [extension, count] : file_count) {
            std::cout << extension << ": " << count << " files" << std::endl;
        }
    }

    /*
    Metody sluzace do przechwytywania i przetwarzania danych wejsciowych
    */
    std::vector<fs::path> user_input_move_copy() {

        fs::path source, destination;
        std::cout << "Enter source path: ";
        std::cin >> source;
        std::cout << "Enter destination path: ";
        std::cin >> destination;

        std::vector<fs::path> out = { source, destination };

        return out;
    }

    fs::path user_input_remove_create() {

        fs::path root_path;
        std::string dir_name;
        std::cout << "Enter the root directory path: ";
        std::cin >> root_path;
        std::cout << "Enter the directory name to remove/create: ";
        std::cin >> dir_name;

        fs::path full_path = root_path / dir_name;

        return full_path;
    }

    fs::path user_input_scan() {
        fs::path dir_to_scan;
        std::cout << "Enter the full or relative path to the directory you want to scan: ";
        std::cin >> dir_to_scan;

        if (!fs::exists(dir_to_scan) || !fs::is_directory(dir_to_scan)) {
            std::cerr << "The provided path does not exist or is not a directory." << std::endl;
            exit(1);
        }

        return dir_to_scan;
    }

    std::string user_input_command() {
        std::string command;
        std::cout << "Enter command: ";
        std::cin >> command;
        return command;
    }

    std::vector<std::string> user_input_file_type() {
        std::vector<std::string> file_types;
        std::cout << "Enter file extensions to search for (e.g., .txt, .cpp, .jpg), finish by typing 'end': ";
        std::string ext;
        while (std::cin >> ext && ext != "end") {
            if (ext[0] != '.') {
                ext = "." + ext; // dodaje kropke, jesli uzytkownik zapomni
            }
            file_types.push_back(ext);
        }
        return file_types;
    }

    bool user_input_recursive() {
        std::string choice;
        std::cout << "Do you want to scan recursively? (yes/no): ";
        std::cin >> choice;
        return (choice == "yes" || choice == "y");
    }
};

int main() {
    FileManager fileManager;
    fileManager.run();
    return 0;
}
